import type * as Next from 'next';
import type { CacheTagsRegistry } from './registry/type';
import { defaultGenerateHash } from './hash';
type Config<R extends CacheTagsRegistry> = {
    /**
     * Disable/enable logging of cache-tag actions.
     */
    log?: boolean;
    /**
     * A cache-tag registry/store.
     */
    registry: R;
    /**
     * An indepotent cache-tag hash generator.
     */
    generateHash?: typeof defaultGenerateHash | false;
};
interface TagsResolver {
    (req: Next.NextApiRequest, res: Next.NextApiResponse): Promise<string[]> | string[];
}
type InvalidatorConfig = {
    /**
     * Wheter the invalidator route should wait for it to operate.
     *
     * Defaults to false, and should be used carefully not to exceed lambda execution time limit.
     */
    wait?: boolean;
    /**
     * Cache-tags resolver for an upcoming invalidation call.
     */
    resolver: TagsResolver;
    /**
     * Controls how the API Route should be resolved in case of invalidation success.
     */
    onSuccess?: (req: Parameters<TagsResolver>[0], res: Parameters<TagsResolver>[1], tags: string[]) => Promise<void> | void;
    /**
     * Controls how the API Route should be resolved in case of invalidation error.
     */
    onError?: (error: any, req: Parameters<TagsResolver>[0], res: Parameters<TagsResolver>[1], tags?: string[]) => Promise<void> | void;
};
/**
 * Cache-tags service for Next.js.
 */
declare class CacheTags<R extends CacheTagsRegistry> {
    log: boolean;
    registry: R;
    generateHash: typeof defaultGenerateHash;
    constructor(config: Config<R>);
    /**
     * Predicate for getStaticProps call-site.
     */
    private isGetStaticPropsTrace;
    /**
     * Resolve the current page's pathname from Next.js static props context info.
     */
    private getPagePathname;
    /**
     * Register tags into a Next.js extracted path.
     */
    register(ctx: Next.GetStaticPropsContext, tags: string[]): void | Promise<void>;
    /**
     * Register a tags/path relationship.
     */
    registerPath(path: string, tags: string[]): void | Promise<void>;
    /**
     * Triggers the invalidation of set of tags.
     */
    invalidate(res: Next.NextApiResponse, tags: string[], wait?: boolean): Promise<void>;
    /**
     * Generates a Next.js API Route for cache-tag invalidation.
     */
    invalidator({ wait, resolver, onSuccess, onError, }: InvalidatorConfig): Next.NextApiHandler;
}
export type { TagsResolver };
export { CacheTags };
