"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.CacheTags = void 0;
const tslib_1 = require("tslib");
const router_1 = tslib_1.__importDefault(require("next/router"));
const router_2 = require("next/dist/shared/lib/router/router");
const stack_1 = require("./stack");
const hash_1 = require("./hash");
const regex = {
    pathname: /(pages|app)\/(?<pathname>.+)(\.index)?\.[tj]sx?$/u,
};
/**
 * Cache-tags service for Next.js.
 */
class CacheTags {
    constructor(config) {
        var _a, _b;
        /**
         * Predicate for getStaticProps call-site.
         */
        this.isGetStaticPropsTrace = (callSite) => callSite.getFunctionName() === 'getStaticProps';
        this.log = (_a = config.log) !== null && _a !== void 0 ? _a : false;
        this.registry = config.registry;
        this.generateHash =
            config.generateHash === false
                ? hash_1.noHash
                : (_b = config.generateHash) !== null && _b !== void 0 ? _b : hash_1.defaultGenerateHash;
    }
    /**
     * Resolve the current page's pathname from Next.js static props context info.
     */
    getPagePathname(ctx) {
        var _a, _b, _c;
        const stack = (0, stack_1.get)();
        const trace = stack.find(this.isGetStaticPropsTrace);
        const path = (_a = trace === null || trace === void 0 ? void 0 : trace.getFileName()) !== null && _a !== void 0 ? _a : trace === null || trace === void 0 ? void 0 : trace.getEvalOrigin();
        const { pathname } = (_c = (_b = path === null || path === void 0 ? void 0 : path.match(regex.pathname)) === null || _b === void 0 ? void 0 : _b.groups) !== null && _c !== void 0 ? _c : {};
        if (!pathname) {
            throw new Error('Could not resolve page pathname');
        }
        return (0, router_2.resolveHref)(router_1.default, { pathname, query: ctx.params }, true)[1];
    }
    /**
     * Register tags into a Next.js extracted path.
     */
    register(ctx, tags) {
        return this.registerPath(this.getPagePathname(ctx), tags);
    }
    /**
     * Register a tags/path relationship.
     */
    registerPath(path, tags) {
        const hashed = tags.map((tag) => this.generateHash(tag));
        /* istanbul ignore next */
        if (this.log) {
            console.log(`[next-cache-tags] Regenerating ${path}:`);
            console.log(`  - Cache-tags: ${hashed.length}`);
        }
        return this.registry.register(path, hashed);
    }
    /**
     * Triggers the invalidation of set of tags.
     */
    invalidate(res, tags, wait) {
        var _a;
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const invalidating = [];
            for (const tag of tags) {
                this.log && console.log(`[next-cache-tags] Invalidating "${tag}":`);
                // Fetch the paths related to the invalidating cache-tag.
                const paths = (_a = (yield this.registry.extract(this.generateHash(tag)))) !== null && _a !== void 0 ? _a : [];
                for (const path of paths) {
                    this.log && console.log(`  - ${path}`);
                    // Dispatch revalidation.
                    invalidating.push(res.revalidate(path));
                }
            }
            /* istanbul ignore next */
            if (wait) {
                yield Promise.allSettled(invalidating);
            }
        });
    }
    /**
     * Generates a Next.js API Route for cache-tag invalidation.
     */
    invalidator({ wait = false, resolver, onSuccess = (_req, res) => res.send('ok'), onError = (error, _req, res) => {
        console.error(error);
        res.status(500).json({
            message: error instanceof Error ? error.message : 'Unknown error',
        });
    }, }) {
        return (req, res) => tslib_1.__awaiter(this, void 0, void 0, function* () {
            let tags = undefined;
            try {
                // Retrieves tags
                const tags = yield resolver(req, res);
                // Perform invalidation.
                yield this.invalidate(res, tags, wait);
                yield onSuccess(req, res, tags);
            }
            catch (err) {
                yield onError(err, req, res, tags);
            }
        });
    }
}
exports.CacheTags = CacheTags;
//# sourceMappingURL=cache-tags.js.map